@@ -1,5 +1,6 @@
 from io import BytesIO
 import zlib
+import sys
 
 from warcio.utils import BUFF_SIZE
 
@@ -62,7 +63,9 @@ class BufferedReader(object):
 
     def __init__(self, stream, block_size=BUFF_SIZE,
                  decomp_type=None,
-                 starting_data=None):
+                 starting_data=None,
+                 read_all_members=False):
+
         self.stream = stream
         self.block_size = block_size
 
@@ -73,6 +76,8 @@ class BufferedReader(object):
         self.num_read = 0
         self.buff_size = 0
 
+        self.read_all_members = read_all_members
+
     def set_decomp(self, decomp_type):
         self._init_decomp(decomp_type)
 
@@ -110,7 +115,7 @@ class BufferedReader(object):
         # if raw data is not empty and decompressor set, but
         # decompressed buff is empty, keep reading --
         # decompressor likely needs more data to decompress
-        while data and self.decompressor and self.empty():
+        while data and self.decompressor and not self.decompressor.unused_data and self.empty():
             data = self.stream.read(block_size)
             self._process_read(data)
 
@@ -140,7 +145,7 @@ class BufferedReader(object):
                         self.decompressor = None
                 # otherwise (partly decompressed), something is wrong
                 else:
-                    print(str(e))
+                    sys.stderr.write(str(e) + '\n')
                     return b''
         return data
 
@@ -200,8 +205,18 @@ class BufferedReader(object):
 
         return linebuff
 
+    def tell(self):
+        return self.num_read
+
     def empty(self):
-        return not self.buff or self.buff.tell() >= self.buff_size
+        if not self.buff or self.buff.tell() >= self.buff_size:
+            # if reading all members, attempt to get next member automatically
+            if self.read_all_members:
+                self.read_next_member()
+
+            return True
+
+        return False
 
     def read_next_member(self):
         if not self.decompressor or not self.decompressor.unused_data:
@@ -225,6 +240,15 @@ class BufferedReader(object):
             self.stream.close()
             self.stream = None
 
+        self.buff = None
+
+        self.close_decompressor()
+
+    def close_decompressor(self):
+        if self.decompressor:
+            self.decompressor.flush()
+            self.decompressor = None
+
     @classmethod
     def get_supported_decompressors(cls):
         return cls.DECOMPRESSORS.keys()
@@ -298,13 +322,23 @@ class ChunkedDataReader(BufferedReader):
     def _try_decode(self, length_header):
         # decode length header
         try:
-            chunk_size = int(length_header.strip().split(b';')[0], 16)
-        except ValueError:
+            # ensure line ends with \r\n
+            assert(length_header[-2:] == b'\r\n')
+            chunk_size = length_header[:-2].split(b';')[0]
+            chunk_size = int(chunk_size, 16)
+            # sanity check chunk size
+            assert(chunk_size <= 2**31)
+        except (ValueError, AssertionError):
             raise ChunkedDataException(b"Couldn't decode length header " +
                                        length_header)
 
         if not chunk_size:
-            # chunk_size 0 indicates end of file
+            # chunk_size 0 indicates end of file. read final bytes to compute digest.
+            final_data = self.stream.read(2)
+            try:
+                assert(final_data == b'\r\n')
+            except AssertionError:
+                raise ChunkedDataException(b"Incorrect \r\n after length header of 0")
             self.all_chunks_read = True
             self._process_read(b'')
             return
